/*
 *  TOPPERS/FI4 Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Fullset uItron4 Kernel
 * 
 *  Copyright (C) 2003-2004 by Monami software Limited Partnership, JAPAN
 * 
 *  L쌠҂́Cȉ (1)`(4) ̏CFree Software Foundation 
 *  ɂČ\Ă GNU General Public License  Version 2 ɋL
 *  qĂ𖞂ꍇɌC{\tgEFAi{\tgEFA
 *  ς̂܂ށDȉjgpEEρEĔzziȉC
 *  pƌĂԁj邱Ƃ𖳏ŋD
 *  (1) {\tgEFA\[XR[ȟ`ŗpꍇɂ́CL̒
 *      \C̗pщL̖ۏ؋K肪Ĉ܂܂̌`Ń\[
 *      XR[hɊ܂܂Ă邱ƁD
 *  (2) {\tgEFACCu`ȂǁC̃\tgEFAJɎg
 *      pł`ōĔzzꍇɂ́CĔzzɔhLgip
 *      ҃}jAȂǁjɁCL̒쌠\C̗pщL
 *      ̖ۏ؋Kfڂ邱ƁD
 *  (3) {\tgEFAC@ɑgݍނȂǁC̃\tgEFAJɎg
 *      płȂ`ōĔzzꍇɂ́Ĉꂩ̏𖞂
 *      ƁD
 *    (a) ĔzzɔhLgip҃}jAȂǁjɁCL̒
 *        쌠\C̗pщL̖ۏ؋Kfڂ邱ƁD
 *    (b) Ĕzž`ԂCʂɒ߂@ɂāCTOPPERSvWFNg
 *        񍐂邱ƁD
 *  (4) {\tgEFA̗pɂ蒼ړI܂͊ԐړIɐ邢Ȃ鑹
 *      QCL쌠҂TOPPERSvWFNgƐӂ邱ƁD
 * 
 *  {\tgEFÁCۏ؂Œ񋟂Ă̂łDL쌠҂
 *  TOPPERSvWFNǵC{\tgEFAɊւāC̓Kp\
 *  ܂߂āCȂۏ؂sȂD܂C{\tgEFA̗pɂ蒼
 *  ړI܂͊ԐړIɐȂ鑹QɊւĂC̐ӔC𕉂ȂD
 * 
 *  @(#) $Id: /local/fi4-1.4/toppers/RELEASE_20061105_1/kernel/kmem.c 2750 2006-03-07T16:15:24.532712Z monaka  $
 */

/*
 *  J[lAP[^
 */
#include <limits.h>
#include <stddef.h>
#include "fi4_kernel.h"
#include "kmem.h"

#define alignof(type) offsetof(struct { char a; type b; }, b)

typedef union {
	long l;
#ifdef _int64_
	_int64_ i64;
#endif
#if (__STDC_VERSION__ >= 199901L)
	long long ll;
#endif
	double d;
	long double ld;
} __dummy_for_align;

static const int maxalign = alignof(__dummy_for_align);

/*
 *
 */
#ifdef __kmemsetup

BOOL 
kmem_setup(VP buf, SIZE bufsz, KMEMB **p_kmemb)
{
	KMEMB *top;
	KMEMB *kmemb;

	if (bufsz < sizeof(KMEMB) * 2 ||
	    (UINT)buf % alignof(__dummy_for_align) != 0) {
		return FALSE;
	}

	top = (KMEMB *)buf;
	top->magic = 0;
	top->used = TRUE;
	top->size = 0;
	queue_initialize(&(top->queue));

	kmemb = top + 1;
	kmemb->magic = KMEM_MAGIC;
	kmemb->used = FALSE;
	kmemb->size = bufsz - sizeof(KMEMB);
	queue_insert_prev((QUEUE *)top, (QUEUE *)kmemb);

	*p_kmemb = top;

	return TRUE;
}

#endif

/*
 *
 */
#ifdef __kmemallocp

BOOL kmem_allocatable(KMEMB *kmemb, UINT size)
{
	KMEMB *ptr;

	/* alignmentƂĂB */
	size = ((size + sizeof(KMEMB)) + sizeof(__dummy_for_align) - 1) & - sizeof(__dummy_for_align);

	for (ptr = kmemb; ptr != kmemb; ptr = (KMEMB *)(ptr->queue.next)) {
		if (ptr->magic != KMEM_MAGIC) {
			return FALSE;
		}

		if (ptr->used == FALSE && ptr->size >= size) {
			/*  */
			return TRUE;
		}
	}

	return FALSE;
}

#endif

/*
 *
 */
#ifdef __kmemallocate

BOOL kmem_allocate(KMEMB *kmemb, UINT size, VP *p_blk)
{
	KMEMB *ptr;

	/* alignmentƂĂB */
	size = ((size + sizeof(KMEMB)) + sizeof(__dummy_for_align) - 1) & - sizeof(__dummy_for_align);

	/* ̈̒T */
	for (ptr = (KMEMB *)(kmemb->queue.next); ptr != kmemb; ptr = (KMEMB *)(ptr->queue.next)) {
		if (ptr->magic != KMEM_MAGIC) {
			return FALSE;
		}

		if (ptr->used == FALSE && ptr->size >= size) {
			/*  */

			if (ptr->size - size > KMEM_DELTA) {
				/* ̈ɏ\ȗ]T̂ŁAB*/
				KMEMB *s;
				s = (KMEMB *)((UB *)ptr + size);
				s->size = ptr->size - size;
				s->magic = ptr->magic = KMEM_MAGIC;
				s->used = FALSE;

				ptr->size = size;
				ptr->used = TRUE;

				queue_insert_prev((QUEUE *)ptr, (QUEUE *)s);

			} else {
				/* ̈ɏ\ȗ]T̂ŁASĂLB*/
				ptr->used = TRUE;
			}

			/* [U̂́Awb_̒ォB*/
			*p_blk = (VP)(ptr + 1);
			return TRUE;
		}
	}

	return FALSE;
}

#endif

/*
 *
 */
#ifdef __kmemrelease

BOOL kmem_release(KMEMB *kmemb, VP buf)
{
	KMEMB *hdr;
	KMEMB *check_kmemb;

	if ((KMEMB *)buf < kmemb) {
		/*
		 *  ̏͂蓾ȂB
		 */
		return FALSE;
	}
	hdr = (KMEMB *)buf - 1;

	if (hdr->magic != KMEM_MAGIC) {
		/*
		 * }WbNioق΁A
		 * sȃAhXnėƂ݂ȂB
		 */
		return FALSE;
	}

	hdr->used = FALSE;
/* Toppers porting F.Okazaki Delete start@magicƓxƎgpłȂ */
/*	hdr->magic = 0;*/
/* Toppers porting F.Okazaki Delete end */

/* Toppers porting F.Okazaki Delete,Append start@̂Ƌtł*/
/*	check_kmemb = (KMEMB *)(hdr->queue.next);*/
	check_kmemb = (KMEMB *)(hdr->queue.prev);
/* Toppers porting F.Okazaki Delete,Append end */
	if (check_kmemb->used == FALSE) {
		hdr->size += check_kmemb->size;
		check_kmemb->magic = 0;
		check_kmemb->queue.next->prev = check_kmemb->queue.prev;
		check_kmemb->queue.prev->next = check_kmemb->queue.next;
	}

	/*
	 *  ÖȂAB
	 */
/* Toppers porting F.Okazaki Delete,Append start@̂Ƌtł*/
/*	check_kmemb = (KMEMB *)(hdr->queue.prev);*/
	check_kmemb = (KMEMB *)(hdr->queue.next);
/* Toppers porting F.Okazaki Delete,Append end */
	if (check_kmemb->used == FALSE) {
		check_kmemb->size += hdr->size;
		hdr->magic = 0;
		hdr->queue.next->prev = hdr->queue.prev;
		hdr->queue.prev->next = hdr->queue.next;
	}

	return TRUE;
}

#endif

/*
 *
 */
#ifdef __kmemstat

BOOL
kmem_status(KMEMB *kmemb, SIZE *remain, UINT *maxblk)
{
	KMEMB *blk;
	QUEUE *entry = NULL;
	SIZE maxblksize = 0;

	if (kmemb->magic != 0) {
		/*
		 * }WbNioق΁A
		 * sȃAhXnėƂ݂ȂB
		 */
		return FALSE;
	}

	*remain = 0;
	while((entry = queue_enumerate((QUEUE *)kmemb, entry)) != NULL) {
		blk = (KMEMB *)entry;

		if (blk->used == FALSE) {
			*remain += blk->size;
			if (maxblksize < blk->size) {
				maxblksize = blk->size;
			}
		}
	}

	/*
	 * ref_mpl ̎dlɍ킹B
	 * uɊlłő̃ubNTCYUINT^ŕ\킹
	 * ől傫ꍇɂ́AUINT^ŕ\킹őlcBv
	 */
	if (maxblksize >= UINT_MAX) {
		*maxblk = UINT_MAX;
	} else {
		*maxblk = (UINT)maxblksize;
	}
	
	return TRUE;
}
#endif /*__kmemstat*/
